# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/annotations/validation.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/annotations/validation.proto',
  package='Ydb',
  syntax='proto3',
  serialized_options=b'\n\016tech.ydb.protoZ2github.com/ydb-platform/ydb-go-genproto/protos/Ydb\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#protos/annotations/validation.proto\x12\x03Ydb\x1a google/protobuf/descriptor.proto\"\x9b\x01\n\x05Limit\x12!\n\x05range\x18\x01 \x01(\x0b\x32\x10.Ydb.Limit.RangeH\x00\x12\x0c\n\x02lt\x18\x02 \x01(\rH\x00\x12\x0c\n\x02le\x18\x03 \x01(\rH\x00\x12\x0c\n\x02\x65q\x18\x04 \x01(\rH\x00\x12\x0c\n\x02ge\x18\x05 \x01(\rH\x00\x12\x0c\n\x02gt\x18\x06 \x01(\rH\x00\x1a!\n\x05Range\x12\x0b\n\x03min\x18\x01 \x01(\r\x12\x0b\n\x03max\x18\x02 \x01(\rB\x06\n\x04kind\"3\n\x06MapKey\x12\x1a\n\x06length\x18\x01 \x01(\x0b\x32\n.Ydb.Limit\x12\r\n\x05value\x18\x02 \x01(\t:1\n\x08required\x12\x1d.google.protobuf.FieldOptions\x18\xe2\xac\x05 \x01(\x08:9\n\x04size\x12\x1d.google.protobuf.FieldOptions\x18\xe3\xac\x05 \x01(\x0b\x32\n.Ydb.Limit:;\n\x06length\x12\x1d.google.protobuf.FieldOptions\x18\xe4\xac\x05 \x01(\x0b\x32\n.Ydb.Limit:=\n\x07map_key\x12\x1d.google.protobuf.FieldOptions\x18\xe5\xac\x05 \x01(\x0b\x32\x0b.Ydb.MapKey:.\n\x05value\x12\x1d.google.protobuf.FieldOptions\x18\xe6\xac\x05 \x01(\tBG\n\x0etech.ydb.protoZ2github.com/ydb-platform/ydb-go-genproto/protos/Ydb\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])


REQUIRED_FIELD_NUMBER = 87650
required = _descriptor.FieldDescriptor(
  name='required', full_name='Ydb.required', index=0,
  number=87650, type=8, cpp_type=7, label=1,
  has_default_value=False, default_value=False,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
SIZE_FIELD_NUMBER = 87651
size = _descriptor.FieldDescriptor(
  name='size', full_name='Ydb.size', index=1,
  number=87651, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
LENGTH_FIELD_NUMBER = 87652
length = _descriptor.FieldDescriptor(
  name='length', full_name='Ydb.length', index=2,
  number=87652, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
MAP_KEY_FIELD_NUMBER = 87653
map_key = _descriptor.FieldDescriptor(
  name='map_key', full_name='Ydb.map_key', index=3,
  number=87653, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
VALUE_FIELD_NUMBER = 87654
value = _descriptor.FieldDescriptor(
  name='value', full_name='Ydb.value', index=4,
  number=87654, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=b"".decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)


_LIMIT_RANGE = _descriptor.Descriptor(
  name='Range',
  full_name='Ydb.Limit.Range',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='min', full_name='Ydb.Limit.Range.min', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='Ydb.Limit.Range.max', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=193,
  serialized_end=226,
)

_LIMIT = _descriptor.Descriptor(
  name='Limit',
  full_name='Ydb.Limit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='range', full_name='Ydb.Limit.range', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='lt', full_name='Ydb.Limit.lt', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='le', full_name='Ydb.Limit.le', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='eq', full_name='Ydb.Limit.eq', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ge', full_name='Ydb.Limit.ge', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gt', full_name='Ydb.Limit.gt', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LIMIT_RANGE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='Ydb.Limit.kind',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=79,
  serialized_end=234,
)


_MAPKEY = _descriptor.Descriptor(
  name='MapKey',
  full_name='Ydb.MapKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='length', full_name='Ydb.MapKey.length', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Ydb.MapKey.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=236,
  serialized_end=287,
)

_LIMIT_RANGE.containing_type = _LIMIT
_LIMIT.fields_by_name['range'].message_type = _LIMIT_RANGE
_LIMIT.oneofs_by_name['kind'].fields.append(
  _LIMIT.fields_by_name['range'])
_LIMIT.fields_by_name['range'].containing_oneof = _LIMIT.oneofs_by_name['kind']
_LIMIT.oneofs_by_name['kind'].fields.append(
  _LIMIT.fields_by_name['lt'])
_LIMIT.fields_by_name['lt'].containing_oneof = _LIMIT.oneofs_by_name['kind']
_LIMIT.oneofs_by_name['kind'].fields.append(
  _LIMIT.fields_by_name['le'])
_LIMIT.fields_by_name['le'].containing_oneof = _LIMIT.oneofs_by_name['kind']
_LIMIT.oneofs_by_name['kind'].fields.append(
  _LIMIT.fields_by_name['eq'])
_LIMIT.fields_by_name['eq'].containing_oneof = _LIMIT.oneofs_by_name['kind']
_LIMIT.oneofs_by_name['kind'].fields.append(
  _LIMIT.fields_by_name['ge'])
_LIMIT.fields_by_name['ge'].containing_oneof = _LIMIT.oneofs_by_name['kind']
_LIMIT.oneofs_by_name['kind'].fields.append(
  _LIMIT.fields_by_name['gt'])
_LIMIT.fields_by_name['gt'].containing_oneof = _LIMIT.oneofs_by_name['kind']
_MAPKEY.fields_by_name['length'].message_type = _LIMIT
DESCRIPTOR.message_types_by_name['Limit'] = _LIMIT
DESCRIPTOR.message_types_by_name['MapKey'] = _MAPKEY
DESCRIPTOR.extensions_by_name['required'] = required
DESCRIPTOR.extensions_by_name['size'] = size
DESCRIPTOR.extensions_by_name['length'] = length
DESCRIPTOR.extensions_by_name['map_key'] = map_key
DESCRIPTOR.extensions_by_name['value'] = value
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Limit = _reflection.GeneratedProtocolMessageType('Limit', (_message.Message,), {

  'Range' : _reflection.GeneratedProtocolMessageType('Range', (_message.Message,), {
    'DESCRIPTOR' : _LIMIT_RANGE,
    '__module__' : 'protos.annotations.validation_pb2'
    # @@protoc_insertion_point(class_scope:Ydb.Limit.Range)
    })
  ,
  'DESCRIPTOR' : _LIMIT,
  '__module__' : 'protos.annotations.validation_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Limit)
  })
_sym_db.RegisterMessage(Limit)
_sym_db.RegisterMessage(Limit.Range)

MapKey = _reflection.GeneratedProtocolMessageType('MapKey', (_message.Message,), {
  'DESCRIPTOR' : _MAPKEY,
  '__module__' : 'protos.annotations.validation_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.MapKey)
  })
_sym_db.RegisterMessage(MapKey)

google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(required)
size.message_type = _LIMIT
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(size)
length.message_type = _LIMIT
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(length)
map_key.message_type = _MAPKEY
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(map_key)
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(value)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from .driver import Driver  # noqa
from .table import SessionPool, retry_operation  # noqa
from .query import QuerySessionPool, QuerySession, QueryTxContext  # noqa
from .coordination import CoordinationClient  # noqa



from __future__ import annotations

import asyncio
from typing import List

import pytest

import ydb.aio


@pytest.mark.asyncio
class TestTopicWriterAsyncIO:
    async def test_send_message(self, driver: ydb.aio.Driver, topic_path):
        writer = driver.topic_client.writer(topic_path, producer_id="test")
        await writer.write(ydb.TopicWriterMessage(data="123".encode()))
        await writer.close()

    async def test_send_message_with_metadata(self, driver: ydb.aio.Driver, topic_path):
        writer = driver.topic_client.writer(topic_path, producer_id="test")
        await writer.write(ydb.TopicWriterMessage(data="123".encode(), metadata_items={"key": "value"}))
        await writer.close()

    async def test_wait_last_seqno(self, driver: ydb.aio.Driver, topic_path):
        async with driver.topic_client.writer(
            topic_path,
            producer_id="test",
            auto_seqno=False,
        ) as writer:
            ret = await writer.write_with_ack(ydb.TopicWriterMessage(data="123".encode(), seqno=5))
            assert ret.offset == 0

        async with driver.topic_client.writer(
            topic_path,
            producer_id="test",
        ) as writer2:
            init_info = await writer2.wait_init()
            assert init_info.last_seqno == 5

    async def test_link_to_client(self, driver, topic_path, topic_consumer):
        async with driver.topic_client.writer(topic_path) as writer:
            assert writer._parent is driver.topic_client

    async def test_random_producer_id(self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO):
        async with driver.topic_client.writer(topic_path) as writer:
            await writer.write(ydb.TopicWriterMessage(data="123".encode()))
        async with driver.topic_client.writer(topic_path) as writer:
            await writer.write(ydb.TopicWriterMessage(data="123".encode()))

        msg1 = await topic_reader.receive_message()
        msg2 = await topic_reader.receive_message()

        assert msg1.producer_id != msg2.producer_id

    async def test_auto_flush_on_close(self, driver: ydb.aio.Driver, topic_path):
        async with driver.topic_client.writer(
            topic_path,
            producer_id="test",
            auto_seqno=False,
        ) as writer:
            last_seqno = 0
            for i in range(10):
                last_seqno = i + 1
                await writer.write(ydb.TopicWriterMessage(data=f"msg-{i}", seqno=last_seqno))

        async with driver.topic_client.writer(
            topic_path,
            producer_id="test",
        ) as writer:
            init_info = await writer.wait_init()
            assert init_info.last_seqno == last_seqno

    async def test_write_multi_message_with_ack(
        self, driver: ydb.aio.Driver, topic_path, topic_reader: ydb.TopicReaderAsyncIO
    ):
        async with driver.topic_client.writer(topic_path) as writer:
            res1, res2 = await writer.write_with_ack(
                [
                    ydb.TopicWriterMessage(data="123".encode()),
                    ydb.TopicWriterMessage(data="456".encode()),
                ]
            )
            assert res1.offset == 0
            assert res2.offset == 1

        msg1 = await topic_reader.receive_message()
        msg2 = await topic_reader.receive_message()

        assert msg1.offset == 0
        assert msg1.seqno == 1
        assert msg1.data == "123".encode()

        assert msg2.offset == 1
        assert msg2.seqno == 2
        assert msg2.data == "456".encode()

    @pytest.mark.parametrize(
        "codec",
        [
            ydb.TopicCodec.RAW,
            ydb.TopicCodec.GZIP,
            None,
        ],
    )
    async def test_write_encoded(self, driver: ydb.aio.Driver, topic_path: str, codec):
        async with driver.topic_client.writer(topic_path, codec=codec) as writer:
            await writer.write("a" * 1000)
            await writer.write("b" * 1000)
            await writer.write("c" * 1000)

    async def test_create_writer_after_stop(self, driver: ydb.aio.Driver, topic_path: str):
        await driver.stop()
        with pytest.raises(ydb.Error):
            async with driver.topic_client.writer(topic_path) as writer:
                await writer.write_with_ack("123")

    @pytest.mark.skip(reason="something wrong with this test, need to assess")
    async def test_send_message_after_stop(self, driver: ydb.aio.Driver, topic_path: str):
        writer = driver.topic_client.writer(topic_path)
        await driver.stop()
        with pytest.raises(ydb.Error):
            await writer.write_with_ack("123")

    async def test_preserve_exception_on_cm_close(self, driver: ydb.aio.Driver, topic_path: str):
        class TestException(Exception):
            pass

        with pytest.raises(TestException):
            async with driver.topic_client.writer(topic_path) as writer:
                await writer.wait_init()
                await driver.stop()  # will raise exception on topic writer __exit__

                # ensure writer has exception internally
                with pytest.raises((ydb.Error, asyncio.CancelledError)):
                    await writer.write_with_ack("123")

                raise TestException()


class TestTopicWriterSync:
    def test_send_message(self, driver_sync: ydb.Driver, topic_path):
        writer = driver_sync.topic_client.writer(topic_path, producer_id="test")
        writer.write(ydb.TopicWriterMessage(data="123".encode()))
        writer.close()

    def test_send_message_with_metadata(self, driver_sync: ydb.Driver, topic_path):
        writer = driver_sync.topic_client.writer(topic_path, producer_id="test")
        writer.write(ydb.TopicWriterMessage(data="123".encode(), metadata_items={"key": "value"}))
        writer.close()

    def test_wait_last_seqno(self, driver_sync: ydb.Driver, topic_path):
        with driver_sync.topic_client.writer(
            topic_path,
            producer_id="test",
            auto_seqno=False,
        ) as writer:
            writer.write_with_ack(ydb.TopicWriterMessage(data="123".encode(), seqno=5))

        with driver_sync.topic_client.writer(
            topic_path,
            producer_id="test",
        ) as writer2:
            init_info = writer2.wait_init()
            assert init_info.last_seqno == 5

    def test_auto_flush_on_close(self, driver_sync: ydb.Driver, topic_path):
        with driver_sync.topic_client.writer(
            topic_path,
            producer_id="test",
            auto_seqno=False,
        ) as writer:
            last_seqno = 0
            for i in range(10):
                last_seqno = i + 1
                writer.write(ydb.TopicWriterMessage(data=f"msg-{i}", seqno=last_seqno))

        with driver_sync.topic_client.writer(
            topic_path,
            producer_id="test",
        ) as writer:
            init_info = writer.wait_init()
            assert init_info.last_seqno == last_seqno

    def test_link_to_client(self, driver_sync, topic_path, topic_consumer):
        with driver_sync.topic_client.writer(topic_path) as writer:
            assert writer._parent is driver_sync.topic_client

    def test_random_producer_id(
        self,
        driver_sync: ydb.aio.Driver,
        topic_path,
        topic_reader_sync: ydb.TopicReader,
    ):
        with driver_sync.topic_client.writer(topic_path) as writer:
            writer.write(ydb.TopicWriterMessage(data="123".encode()))
        with driver_sync.topic_client.writer(topic_path) as writer:
            writer.write(ydb.TopicWriterMessage(data="123".encode()))

        msg1 = topic_reader_sync.receive_message()
        msg2 = topic_reader_sync.receive_message()

        assert msg1.producer_id != msg2.producer_id

    def test_write_multi_message_with_ack(
        self, driver_sync: ydb.Driver, topic_path, topic_reader_sync: ydb.TopicReader
    ):
        with driver_sync.topic_client.writer(topic_path) as writer:
            writer.write_with_ack(
                [
                    ydb.TopicWriterMessage(data="123".encode()),
                    ydb.TopicWriterMessage(data="456".encode()),
                ]
            )

        msg1 = topic_reader_sync.receive_message()
        msg2 = topic_reader_sync.receive_message()

        assert msg1.offset == 0
        assert msg1.seqno == 1
        assert msg1.data == "123".encode()

        assert msg2.offset == 1
        assert msg2.seqno == 2
        assert msg2.data == "456".encode()

    @pytest.mark.parametrize(
        "codec",
        [
            ydb.TopicCodec.RAW,
            ydb.TopicCodec.GZIP,
            None,
        ],
    )
    def test_write_encoded(self, driver_sync: ydb.Driver, topic_path: str, codec):
        with driver_sync.topic_client.writer(topic_path, codec=codec) as writer:
            writer.write("a" * 1000)
            writer.write("b" * 1000)
            writer.write("c" * 1000)

    def test_start_many_sync_writers_in_parallel(self, driver_sync: ydb.Driver, topic_path):
        target_count = 100
        writers = []  # type: List[ydb.TopicWriter]
        for i in range(target_count):
            writer = driver_sync.topic_client.writer(topic_path)
            writers.append(writer)

        for i, writer in enumerate(writers):
            writer.write(str(i))

        for writer in writers:
            writer.close()

    def test_create_writer_after_stop(self, driver_sync: ydb.Driver, topic_path: str):
        driver_sync.stop()
        with pytest.raises(ydb.Error):
            with driver_sync.topic_client.writer(topic_path) as writer:
                writer.write_with_ack("123")

    @pytest.mark.skip(reason="something wrong with this test, need to assess")
    def test_send_message_after_stop(self, driver_sync: ydb.Driver, topic_path: str):
        writer = driver_sync.topic_client.writer(topic_path)
        driver_sync.stop()
        with pytest.raises(ydb.Error):
            writer.write_with_ack("123")

    def test_preserve_exception_on_cm_close(self, driver_sync: ydb.Driver, topic_path: str):
        class TestException(Exception):
            pass

        with pytest.raises(TestException):
            with driver_sync.topic_client.writer(topic_path) as writer:
                writer.wait_init()
                driver_sync.stop()  # will raise exception on topic writer __exit__

                # ensure writer has exception internally
                with pytest.raises(ydb.Error):
                    writer.write_with_ack("123")

                raise TestException()



import concurrent.futures
import enum
import datetime
from dataclasses import dataclass
from typing import (
    Union,
    Optional,
    List,
    Mapping,
    Callable,
)

from .events import EventHandler
from ..retries import RetrySettings
from .._grpc.grpcwrapper.ydb_topic import StreamReadMessage, OffsetsRange


@dataclass
class PublicTopicSelector:
    path: str
    partitions: Optional[Union[int, List[int]]] = None
    read_from: Optional[datetime.datetime] = None
    max_lag: Optional[datetime.timedelta] = None
    read_offset: Optional[int] = None

    def _to_topic_read_settings(self) -> StreamReadMessage.InitRequest.TopicReadSettings:
        partitions = self.partitions
        if partitions is None:
            partitions = []

        elif not isinstance(partitions, list):
            partitions = [partitions]

        return StreamReadMessage.InitRequest.TopicReadSettings(
            path=self.path,
            partition_ids=partitions,
            max_lag=self.max_lag,
            read_from=self.read_from,
        )


TopicSelectorTypes = Union[str, PublicTopicSelector, List[Union[str, PublicTopicSelector]]]


@dataclass
class PublicReaderSettings:
    consumer: Optional[str]
    topic: TopicSelectorTypes
    buffer_size_bytes: int = 50 * 1024 * 1024
    auto_partitioning_support: bool = True

    decoders: Union[Mapping[int, Callable[[bytes], bytes]], None] = None
    """decoders: map[codec_code] func(encoded_bytes)->decoded_bytes"""

    # decoder_executor, must be set for handle non raw messages
    decoder_executor: Optional[concurrent.futures.Executor] = None
    update_token_interval: Union[int, float] = 3600
    event_handler: Optional[EventHandler] = None

    def __post_init__(self):
        # check possible create init message
        _ = self._init_message()

    def _init_message(self) -> StreamReadMessage.InitRequest:
        if self.consumer is not None and not isinstance(self.consumer, str):
            raise TypeError("Unsupported type for customer field: '%s'" % type(self.consumer))

        if isinstance(self.topic, list):
            selectors = self.topic
        else:
            selectors = [self.topic]

        for index, selector in enumerate(selectors):
            if isinstance(selector, str):
                selectors[index] = PublicTopicSelector(path=selector)
            elif isinstance(selector, PublicTopicSelector):
                pass
            else:
                raise TypeError("Unsupported type for topic field: '%s'" % type(selector))

        return StreamReadMessage.InitRequest(
            topics_read_settings=list(map(PublicTopicSelector._to_topic_read_settings, selectors)),  # type: ignore
            consumer=self.consumer,
            auto_partitioning_support=self.auto_partitioning_support,
        )

    def _retry_settings(self) -> RetrySettings:
        return RetrySettings(idempotent=True, retry_cancelled=True)


class RetryPolicy:
    connection_timeout_sec: float
    overload_timeout_sec: float
    retry_access_denied: bool = False


class CommitResult:
    topic: str
    partition: int
    offset: int
    state: "CommitResult.State"
    details: str  # for humans only, content messages may be change in any time

    class State(enum.Enum):
        UNSENT = 1  # commit didn't send to the server
        SENT = 2  # commit was sent to server, but ack hasn't received
        ACKED = 3  # ack from server is received


class SessionStat:
    path: str
    partition_id: str
    partition_offsets: OffsetsRange
    committed_offset: int
    write_time_high_watermark: datetime.datetime
    write_time_high_watermark_timestamp_nano: int


class StubEvent:
    pass



from ydb.iam.auth import BaseJWTCredentials
from unittest.mock import patch, mock_open

CONTENT1 = '{"service_account_id":"my_sa", "id":"123", "private_key":"pppp", "user_account_id":"ua_id"}'
CONTENT2 = '{"id":"123", "private_key":"pppp", "user_account_id":"ua_id"}'


class FakeAuth(BaseJWTCredentials):
    def __init__(self, account_id, key_id, private_key, iam_endpoint=None, iam_channel_credentials=None):
        self.account_id = account_id
        self.key_id = key_id
        self.private_key = private_key
        self.iam_endpoint = iam_endpoint
        self.iam_channel_credentials = iam_channel_credentials

    def __eq__(self, other):
        return self.__dict__ == other.__dict__ if type(self) == type(other) else False


@patch("builtins.open", new_callable=mock_open, read_data=CONTENT1)
def test_auth_from_file(mock_file):
    r1 = FakeAuth.from_file("my_file.json", iam_endpoint="my_iam_address", iam_channel_credentials="my_creds")
    mock_file.assert_called_with("my_file.json", "r")

    r2 = FakeAuth.from_content(CONTENT1, iam_endpoint="my_iam_address", iam_channel_credentials="my_creds")
    assert r1 == r2
    assert r1.__dict__ == {
        "account_id": "my_sa",
        "key_id": "123",
        "private_key": "pppp",
        "iam_endpoint": "my_iam_address",
        "iam_channel_credentials": "my_creds",
    }

    r3 = FakeAuth.from_content(CONTENT2)

    assert r3.__dict__ == {
        "account_id": "ua_id",
        "key_id": "123",
        "private_key": "pppp",
        "iam_endpoint": None,
        "iam_channel_credentials": None,
    }



import datetime
import enum
from dataclasses import dataclass
from typing import List


@dataclass
class Entry:
    name: str
    owner: str
    type: "Entry.Type"
    effective_permissions: "Permissions"
    permissions: "Permissions"
    size_bytes: int
    created_at: datetime.datetime

    class Type(enum.IntEnum):
        UNSPECIFIED = 0
        DIRECTORY = 1
        TABLE = 2
        PERS_QUEUE_GROUP = 3
        DATABASE = 4
        RTMR_VOLUME = 5
        BLOCK_STORE_VOLUME = 6
        COORDINATION_NODE = 7
        COLUMN_STORE = 12
        COLUMN_TABLE = 13
        SEQUENCE = 15
        REPLICATION = 16
        TOPIC = 17


@dataclass
class Permissions:
    subject: str
    permission_names: List[str]



import grpc.aio
import time

import abc
import logging
from ydb.iam import auth
from .credentials import AbstractExpiringTokenCredentials

logger = logging.getLogger(__name__)

try:
    import jwt
except ImportError:
    jwt = None

try:
    from yandex.cloud.iam.v1 import iam_token_service_pb2_grpc
    from yandex.cloud.iam.v1 import iam_token_service_pb2
except ImportError:
    iam_token_service_pb2_grpc = None
    iam_token_service_pb2 = None

try:
    import aiohttp
except ImportError:
    aiohttp = None


class TokenServiceCredentials(AbstractExpiringTokenCredentials):
    def __init__(self, iam_endpoint=None, iam_channel_credentials=None):
        super(TokenServiceCredentials, self).__init__()
        assert iam_token_service_pb2_grpc is not None, 'run pip install "ydb[yc]" to use service account credentials'
        self._get_token_request_timeout = 10
        self._iam_endpoint = "iam.api.cloud.yandex.net:443" if iam_endpoint is None else iam_endpoint
        self._iam_channel_credentials = {} if iam_channel_credentials is None else iam_channel_credentials

    def _channel_factory(self):
        return grpc.aio.secure_channel(
            self._iam_endpoint,
            grpc.ssl_channel_credentials(**self._iam_channel_credentials),
        )

    @abc.abstractmethod
    def _get_token_request(self):
        pass

    async def _make_token_request(self):
        async with self._channel_factory() as channel:
            stub = iam_token_service_pb2_grpc.IamTokenServiceStub(channel)
            response = await stub.Create(self._get_token_request(), timeout=self._get_token_request_timeout)
            self.logger.debug(str(response))
            expires_in = max(0, response.expires_at.seconds - int(time.time()))
            return {"access_token": response.iam_token, "expires_in": expires_in}


# IamTokenCredentials need for backward compatibility
# Deprecated
IamTokenCredentials = TokenServiceCredentials


class JWTIamCredentials(TokenServiceCredentials, auth.BaseJWTCredentials):
    def __init__(
        self,
        account_id,
        access_key_id,
        private_key,
        iam_endpoint=None,
        iam_channel_credentials=None,
    ):
        TokenServiceCredentials.__init__(self, iam_endpoint, iam_channel_credentials)
        auth.BaseJWTCredentials.__init__(
            self,
            account_id,
            access_key_id,
            private_key,
            auth.YANDEX_CLOUD_JWT_ALGORITHM,
            auth.YANDEX_CLOUD_IAM_TOKEN_SERVICE_URL,
        )

    def _get_token_request(self):
        return iam_token_service_pb2.CreateIamTokenRequest(jwt=self._get_jwt())


class YandexPassportOAuthIamCredentials(TokenServiceCredentials):
    def __init__(
        self,
        yandex_passport_oauth_token,
        iam_endpoint=None,
        iam_channel_credentials=None,
    ):
        self._yandex_passport_oauth_token = yandex_passport_oauth_token
        super(YandexPassportOAuthIamCredentials, self).__init__(iam_endpoint, iam_channel_credentials)

    def _get_token_request(self):
        return iam_token_service_pb2.CreateIamTokenRequest(
            yandex_passport_oauth_token=self._yandex_passport_oauth_token
        )


class MetadataUrlCredentials(AbstractExpiringTokenCredentials):
    def __init__(self, metadata_url=None):
        super(MetadataUrlCredentials, self).__init__()
        assert aiohttp is not None, "Install aiohttp library to use metadata credentials provider"
        self._metadata_url = auth.DEFAULT_METADATA_URL if metadata_url is None else metadata_url
        self.extra_error_message = "Check that metadata service configured properly and application deployed in VM or function at Yandex.Cloud."
        self._tp.submit(self._refresh_token)

    async def _make_token_request(self):
        timeout = aiohttp.ClientTimeout(total=2)
        async with aiohttp.ClientSession(timeout=timeout) as session:
            async with session.get(self._metadata_url, headers={"Metadata-Flavor": "Google"}) as response:
                if not response.ok:
                    self.logger.error("Error while getting token from metadata: %s" % await response.text())
                response.raise_for_status()
                # response from default metadata credentials provider
                # contains text/plain content type.
                return await response.json(content_type=None)


class ServiceAccountCredentials(JWTIamCredentials):
    def __init__(
        self,
        service_account_id,
        access_key_id,
        private_key,
        iam_endpoint=None,
        iam_channel_credentials=None,
    ):
        super(ServiceAccountCredentials, self).__init__(
            service_account_id,
            access_key_id,
            private_key,
            iam_endpoint,
            iam_channel_credentials,
        )



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor
SENSITIVE_FIELD_NUMBER: _ClassVar[int]
sensitive: _descriptor.FieldDescriptor



from .. import _utilities


class AsyncResponseContextIterator(_utilities.AsyncResponseIterator):
    def __init__(self, it, wrapper, on_error=None):
        super().__init__(it, wrapper)
        self._on_error = on_error

    async def __aenter__(self) -> "AsyncResponseContextIterator":
        return self

    async def _next(self):
        try:
            return await super()._next()
        except Exception as e:
            if self._on_error:
                self._on_error(e)
            raise e

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        #  To close stream on YDB it is necessary to scroll through it to the end
        async for _ in self:
            pass



from unittest import mock

import grpc
import pytest
from grpc._cython import cygrpc

from ydb.aio.query.session import QuerySession
from ydb.aio.query.pool import QuerySessionPool


@pytest.fixture
async def session(driver):
    session = QuerySession(driver)

    yield session

    try:
        await session.delete()
    except BaseException:
        pass


@pytest.fixture
async def tx(session):
    await session.create()
    transaction = session.transaction()

    yield transaction

    try:
        await transaction.rollback()
    except BaseException:
        pass


@pytest.fixture
async def pool(driver):
    async with QuerySessionPool(driver) as pool:
        yield pool


@pytest.fixture
async def ydb_terminates_streams_with_unavailable():
    async def _patch(self):
        message = await self._read()  # Read the first message
        while message is not cygrpc.EOF:  # While the message is not empty, continue reading the stream
            yield message
            message = await self._read()

        # Emulate stream termination
        raise grpc.aio.AioRpcError(
            code=grpc.StatusCode.UNAVAILABLE,
            initial_metadata=await self.initial_metadata(),
            trailing_metadata=await self.trailing_metadata(),
        )

    with mock.patch.object(grpc.aio._call._StreamResponseMixin, "_fetch_stream_responses", _patch):
        yield



from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ListEndpointsRequest(_message.Message):
    __slots__ = ("database", "service")
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    SERVICE_FIELD_NUMBER: _ClassVar[int]
    database: str
    service: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, database: _Optional[str] = ..., service: _Optional[_Iterable[str]] = ...) -> None: ...

class EndpointInfo(_message.Message):
    __slots__ = ("address", "port", "load_factor", "ssl", "service", "location", "node_id", "ip_v4", "ip_v6", "ssl_target_name_override")
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    LOAD_FACTOR_FIELD_NUMBER: _ClassVar[int]
    SSL_FIELD_NUMBER: _ClassVar[int]
    SERVICE_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    IP_V4_FIELD_NUMBER: _ClassVar[int]
    IP_V6_FIELD_NUMBER: _ClassVar[int]
    SSL_TARGET_NAME_OVERRIDE_FIELD_NUMBER: _ClassVar[int]
    address: str
    port: int
    load_factor: float
    ssl: bool
    service: _containers.RepeatedScalarFieldContainer[str]
    location: str
    node_id: int
    ip_v4: _containers.RepeatedScalarFieldContainer[str]
    ip_v6: _containers.RepeatedScalarFieldContainer[str]
    ssl_target_name_override: str
    def __init__(self, address: _Optional[str] = ..., port: _Optional[int] = ..., load_factor: _Optional[float] = ..., ssl: bool = ..., service: _Optional[_Iterable[str]] = ..., location: _Optional[str] = ..., node_id: _Optional[int] = ..., ip_v4: _Optional[_Iterable[str]] = ..., ip_v6: _Optional[_Iterable[str]] = ..., ssl_target_name_override: _Optional[str] = ...) -> None: ...

class ListEndpointsResult(_message.Message):
    __slots__ = ("endpoints", "self_location")
    ENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    SELF_LOCATION_FIELD_NUMBER: _ClassVar[int]
    endpoints: _containers.RepeatedCompositeFieldContainer[EndpointInfo]
    self_location: str
    def __init__(self, endpoints: _Optional[_Iterable[_Union[EndpointInfo, _Mapping]]] = ..., self_location: _Optional[str] = ...) -> None: ...

class ListEndpointsResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class WhoAmIRequest(_message.Message):
    __slots__ = ("include_groups",)
    INCLUDE_GROUPS_FIELD_NUMBER: _ClassVar[int]
    include_groups: bool
    def __init__(self, include_groups: bool = ...) -> None: ...

class WhoAmIResult(_message.Message):
    __slots__ = ("user", "groups")
    USER_FIELD_NUMBER: _ClassVar[int]
    GROUPS_FIELD_NUMBER: _ClassVar[int]
    user: str
    groups: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, user: _Optional[str] = ..., groups: _Optional[_Iterable[str]] = ...) -> None: ...

class WhoAmIResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class NodeLocation(_message.Message):
    __slots__ = ("data_center_num", "room_num", "rack_num", "body_num", "body", "data_center", "module", "rack", "unit")
    DATA_CENTER_NUM_FIELD_NUMBER: _ClassVar[int]
    ROOM_NUM_FIELD_NUMBER: _ClassVar[int]
    RACK_NUM_FIELD_NUMBER: _ClassVar[int]
    BODY_NUM_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    DATA_CENTER_FIELD_NUMBER: _ClassVar[int]
    MODULE_FIELD_NUMBER: _ClassVar[int]
    RACK_FIELD_NUMBER: _ClassVar[int]
    UNIT_FIELD_NUMBER: _ClassVar[int]
    data_center_num: int
    room_num: int
    rack_num: int
    body_num: int
    body: int
    data_center: str
    module: str
    rack: str
    unit: str
    def __init__(self, data_center_num: _Optional[int] = ..., room_num: _Optional[int] = ..., rack_num: _Optional[int] = ..., body_num: _Optional[int] = ..., body: _Optional[int] = ..., data_center: _Optional[str] = ..., module: _Optional[str] = ..., rack: _Optional[str] = ..., unit: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: draft/protos/ydb_object_storage.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_issue_message_pb2 as protos_dot_ydb__issue__message__pb2
from ydb._grpc.v3.protos import ydb_value_pb2 as protos_dot_ydb__value__pb2
from ydb._grpc.v3.protos import ydb_status_codes_pb2 as protos_dot_ydb__status__codes__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='draft/protos/ydb_object_storage.proto',
  package='Ydb.ObjectStorage',
  syntax='proto3',
  serialized_options=b'\n#tech.ydb.proto.draft.object_storageZFgithub.com/ydb-platform/ydb-go-genproto/draft/protos/Ydb_ObjectStorage\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%draft/protos/ydb_object_storage.proto\x12\x11Ydb.ObjectStorage\x1a\x1eprotos/ydb_issue_message.proto\x1a\x16protos/ydb_value.proto\x1a\x1dprotos/ydb_status_codes.proto\"\xd6\x02\n\x0eListingRequest\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12#\n\nkey_prefix\x18\x02 \x01(\x0b\x32\x0f.Ydb.TypedValue\x12\x1a\n\x12path_column_prefix\x18\x03 \x01(\t\x12\x1d\n\x15path_column_delimiter\x18\x04 \x01(\t\x12\x1a\n\x12\x63ontinuation_token\x18\x05 \x01(\x0c\x12/\n\x16start_after_key_suffix\x18\x06 \x01(\x0b\x32\x0f.Ydb.TypedValue\x12\x10\n\x08max_keys\x18\x07 \x01(\x05\x12\x19\n\x11\x63olumns_to_return\x18\x08 \x03(\t\x12(\n\x0fmatching_filter\x18\n \x01(\x0b\x32\x0f.Ydb.TypedValue\"&\n\nEMatchType\x12\t\n\x05\x45QUAL\x10\x00\x12\r\n\tNOT_EQUAL\x10\x01J\x04\x08\t\x10\n\"\xd7\x01\n\x0fListingResponse\x12)\n\x06status\x18\x01 \x01(\x0e\x32\x19.Ydb.StatusIds.StatusCode\x12\'\n\x06issues\x18\x02 \x03(\x0b\x32\x17.Ydb.Issue.IssueMessage\x12\x17\n\x0f\x63ommon_prefixes\x18\x03 \x03(\t\x12 \n\x08\x63ontents\x18\x04 \x01(\x0b\x32\x0e.Ydb.ResultSet\x12\x14\n\x0cis_truncated\x18\x05 \x01(\x08\x12\x1f\n\x17next_continuation_token\x18\x06 \x01(\x0c\x42p\n#tech.ydb.proto.draft.object_storageZFgithub.com/ydb-platform/ydb-go-genproto/draft/protos/Ydb_ObjectStorage\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[protos_dot_ydb__issue__message__pb2.DESCRIPTOR,protos_dot_ydb__value__pb2.DESCRIPTOR,protos_dot_ydb__status__codes__pb2.DESCRIPTOR,])



_LISTINGREQUEST_EMATCHTYPE = _descriptor.EnumDescriptor(
  name='EMatchType',
  full_name='Ydb.ObjectStorage.ListingRequest.EMatchType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EQUAL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NOT_EQUAL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=446,
  serialized_end=484,
)
_sym_db.RegisterEnumDescriptor(_LISTINGREQUEST_EMATCHTYPE)


_LISTINGREQUEST = _descriptor.Descriptor(
  name='ListingRequest',
  full_name='Ydb.ObjectStorage.ListingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='table_name', full_name='Ydb.ObjectStorage.ListingRequest.table_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_prefix', full_name='Ydb.ObjectStorage.ListingRequest.key_prefix', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path_column_prefix', full_name='Ydb.ObjectStorage.ListingRequest.path_column_prefix', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path_column_delimiter', full_name='Ydb.ObjectStorage.ListingRequest.path_column_delimiter', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='continuation_token', full_name='Ydb.ObjectStorage.ListingRequest.continuation_token', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_after_key_suffix', full_name='Ydb.ObjectStorage.ListingRequest.start_after_key_suffix', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_keys', full_name='Ydb.ObjectStorage.ListingRequest.max_keys', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='columns_to_return', full_name='Ydb.ObjectStorage.ListingRequest.columns_to_return', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='matching_filter', full_name='Ydb.ObjectStorage.ListingRequest.matching_filter', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LISTINGREQUEST_EMATCHTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=148,
  serialized_end=490,
)


_LISTINGRESPONSE = _descriptor.Descriptor(
  name='ListingResponse',
  full_name='Ydb.ObjectStorage.ListingResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='Ydb.ObjectStorage.ListingResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='issues', full_name='Ydb.ObjectStorage.ListingResponse.issues', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='common_prefixes', full_name='Ydb.ObjectStorage.ListingResponse.common_prefixes', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='contents', full_name='Ydb.ObjectStorage.ListingResponse.contents', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_truncated', full_name='Ydb.ObjectStorage.ListingResponse.is_truncated', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_continuation_token', full_name='Ydb.ObjectStorage.ListingResponse.next_continuation_token', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=493,
  serialized_end=708,
)

_LISTINGREQUEST.fields_by_name['key_prefix'].message_type = protos_dot_ydb__value__pb2._TYPEDVALUE
_LISTINGREQUEST.fields_by_name['start_after_key_suffix'].message_type = protos_dot_ydb__value__pb2._TYPEDVALUE
_LISTINGREQUEST.fields_by_name['matching_filter'].message_type = protos_dot_ydb__value__pb2._TYPEDVALUE
_LISTINGREQUEST_EMATCHTYPE.containing_type = _LISTINGREQUEST
_LISTINGRESPONSE.fields_by_name['status'].enum_type = protos_dot_ydb__status__codes__pb2._STATUSIDS_STATUSCODE
_LISTINGRESPONSE.fields_by_name['issues'].message_type = protos_dot_ydb__issue__message__pb2._ISSUEMESSAGE
_LISTINGRESPONSE.fields_by_name['contents'].message_type = protos_dot_ydb__value__pb2._RESULTSET
DESCRIPTOR.message_types_by_name['ListingRequest'] = _LISTINGREQUEST
DESCRIPTOR.message_types_by_name['ListingResponse'] = _LISTINGRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListingRequest = _reflection.GeneratedProtocolMessageType('ListingRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINGREQUEST,
  '__module__' : 'draft.protos.ydb_object_storage_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.ObjectStorage.ListingRequest)
  })
_sym_db.RegisterMessage(ListingRequest)

ListingResponse = _reflection.GeneratedProtocolMessageType('ListingResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINGRESPONSE,
  '__module__' : 'draft.protos.ydb_object_storage_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.ObjectStorage.ListingResponse)
  })
_sym_db.RegisterMessage(ListingResponse)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



__all__ = [
    "BaseQueryTxMode",
    "QueryExplainResultFormat",
    "QueryOnlineReadOnly",
    "QuerySerializableReadWrite",
    "QuerySnapshotReadOnly",
    "QuerySnapshotReadWrite",
    "QueryStaleReadOnly",
    "QuerySessionPool",
    "QueryClientSettings",
    "QuerySession",
    "QueryStatsMode",
    "QueryTxContext",
    "QuerySchemaInclusionMode",
    "QueryResultSetFormat",
    "ArrowCompressionCodecType",
    "ArrowCompressionCodec",
    "ArrowFormatSettings",
    "ArrowFormatMeta",
]

import logging

from .base import (
    QueryClientSettings,
    QueryExplainResultFormat,
    QueryStatsMode,
    QuerySchemaInclusionMode,
    QueryResultSetFormat,
)

from .session import QuerySession
from .transaction import QueryTxContext

from .._grpc.grpcwrapper import common_utils
from .._grpc.grpcwrapper.ydb_query_public_types import (
    BaseQueryTxMode,
    QueryOnlineReadOnly,
    QuerySerializableReadWrite,
    QuerySnapshotReadOnly,
    QuerySnapshotReadWrite,
    QueryStaleReadOnly,
    ArrowCompressionCodecType,
    ArrowCompressionCodec,
    ArrowFormatSettings,
    ArrowFormatMeta,
)

from .pool import QuerySessionPool

logger = logging.getLogger(__name__)


class QueryClientSync:
    def __init__(self, driver: common_utils.SupportedDriverType, query_client_settings: QueryClientSettings = None):
        self._driver = driver
        self._settings = query_client_settings

    def session(self) -> QuerySession:
        return QuerySession(self._driver, self._settings)



from __future__ import annotations

import asyncio
import logging
from typing import Dict

from ... import issues
from ..._grpc.grpcwrapper.common_utils import IToProto
from ..._grpc.grpcwrapper.ydb_coordination import FromServer
from .stream import CoordinationStream

logger = logging.getLogger(__name__)


class CoordinationReconnector:
    def __init__(self, driver, node_path: str, timeout_millis: int = 30000):
        self._driver = driver
        self._node_path = node_path
        self._timeout_millis = timeout_millis
        self._wait_timeout = timeout_millis / 1000

        self._stream = None
        self._session_id = None

        self._pending_futures: Dict[int, asyncio.Future] = {}
        self._pending_requests: Dict[int, IToProto] = {}

        self._send_lock = asyncio.Lock()
        self._connection_task = None
        self._closed = False

    async def stop(self):
        self._closed = True

        if self._connection_task:
            self._connection_task.cancel()
            try:
                await self._connection_task
            except asyncio.CancelledError:
                pass

        if self._stream:
            await self._stream.close()

        for fut in self._pending_futures.values():
            if not fut.done():
                fut.set_exception(asyncio.CancelledError())

        self._pending_futures.clear()
        self._pending_requests.clear()

    async def send_and_wait(self, req: IToProto):
        if self._closed:
            raise issues.Error("Reconnector closed")

        if self._connection_task is None:
            self._connection_task = asyncio.create_task(self._connection_loop())

        while not self._stream or self._stream._closed:
            await asyncio.sleep(0)

        req_id = getattr(req, "req_id")
        loop = asyncio.get_running_loop()
        fut = loop.create_future()

        self._pending_futures[req_id] = fut
        self._pending_requests[req_id] = req

        async with self._send_lock:
            await self._stream.send(req)

        return await asyncio.wait_for(fut, self._wait_timeout)

    async def _connection_loop(self):
        while not self._closed:
            try:
                stream = CoordinationStream(self._driver)
                await stream.start_session(
                    self._node_path,
                    self._timeout_millis,
                    session_id=self._session_id,
                )

                self._stream = stream
                self._session_id = stream.session_id

                for req in self._pending_requests.values():
                    await stream.send(req)

                await self._dispatch_loop(stream)

            except asyncio.CancelledError:
                return
            except Exception as exc:
                logger.debug("Coordination stream error: %r", exc)
            finally:
                if self._stream:
                    await self._stream.close()
                    self._stream = None

    async def _dispatch_loop(self, stream):
        while not self._closed and self._stream is stream:
            resp = await stream.receive(self._wait_timeout)
            if not resp:
                continue

            fs = FromServer.from_proto(resp)
            payload = next(
                (
                    getattr(fs, name)
                    for name in (
                        "acquire_semaphore_result",
                        "release_semaphore_result",
                        "describe_semaphore_result",
                        "create_semaphore_result",
                        "update_semaphore_result",
                        "delete_semaphore_result",
                    )
                    if fs.raw.HasField(name)
                ),
                None,
            )

            if not payload:
                continue

            fut = self._pending_futures.pop(payload.req_id, None)
            self._pending_requests.pop(payload.req_id, None)

            if fut and not fut.done():
                fut.set_result(payload)



import pytest

from ydb._utilities import check_module_exists


@pytest.mark.parametrize(
    "path,result",
    [("sys", True), ("ydb", True), ("ydb.some.module.unexisted.test", False)],
)
def test_check_module_exists(path, result):
    assert check_module_exists(path) == result



import abc
from abc import abstractmethod
from . import _apis
from .. import issues, operation


class IDynamicConfigClient(abc.ABC):
    @abstractmethod
    def __init__(self, driver):
        pass

    @abstractmethod
    def replace_config(self, config, dry_run, allow_unknown_fields, settings):
        pass

    @abstractmethod
    def set_config(self, config, dry_run, allow_unknown_fields, settings):
        pass

    @abstractmethod
    def get_config(self, settings):
        pass

    @abstractmethod
    def get_node_labels(self, node_id, settings):
        pass


class DynamicConfig(object):
    __slots__ = ("version", "cluster", "config")

    def __init__(self, version, cluster, config, *args, **kwargs):
        self.version = version
        self.cluster = cluster
        self.config = config


class NodeLabels(object):
    __slots__ = "labels"

    def __init__(self, labels, *args, **kwargs):
        self.labels = labels


def _replace_config_request_factory(config, dry_run, allow_unknown_fields):
    request = _apis.ydb_dynamic_config.ReplaceConfigRequest()
    request.config = config
    request.dry_run = dry_run
    request.allow_unknown_fields = allow_unknown_fields
    return request


def _set_config_request_factory(config, dry_run, allow_unknown_fields):
    request = _apis.ydb_dynamic_config.SetConfigRequest()
    request.config = config
    request.dry_run = dry_run
    request.allow_unknown_fields = allow_unknown_fields
    return request


def _get_config_request_factory():
    request = _apis.ydb_dynamic_config.GetConfigRequest()
    return request


def _get_node_labels_request_factory(node_id):
    request = _apis.ydb_dynamic_config.GetNodeLabelsRequest()
    request.node_id = node_id
    return request


def _wrap_dynamic_config(config_pb, dynamic_config_cls=None, *args, **kwargs):
    dynamic_config_cls = DynamicConfig if dynamic_config_cls is None else dynamic_config_cls
    return dynamic_config_cls(
        config_pb.identity[0].version, config_pb.identity[0].cluster, config_pb.config[0], *args, **kwargs
    )


def _wrap_get_config_response(rpc_state, response):
    issues._process_response(response.operation)
    message = _apis.ydb_dynamic_config.GetConfigResult()
    response.operation.result.Unpack(message)
    return _wrap_dynamic_config(message)


def _wrap_node_labels(labels_pb, node_labels_cls=None, *args, **kwargs):
    node_labels_cls = NodeLabels if node_labels_cls is None else node_labels_cls
    return node_labels_cls(dict([(entry.label, entry.value) for entry in labels_pb.labels]), *args, **kwargs)


def _wrap_get_node_labels_response(rpc_state, response):
    issues._process_response(response.operation)
    message = _apis.ydb_dynamic_config.GetNodeLabelsResult()
    response.operation.result.Unpack(message)
    return _wrap_node_labels(message)


class BaseDynamicConfigClient(IDynamicConfigClient):
    __slots__ = ("_driver",)

    def __init__(self, driver):
        self._driver = driver

    def replace_config(self, config, dry_run, allow_unknown_fields, settings=None):
        return self._driver(
            _replace_config_request_factory(config, dry_run, allow_unknown_fields),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.ReplaceConfig,
            operation.Operation,
            settings,
        )

    def set_config(self, config, dry_run, allow_unknown_fields, settings=None):
        return self._driver(
            _set_config_request_factory(config, dry_run, allow_unknown_fields),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.SetConfig,
            operation.Operation,
            settings,
        )

    def get_config(self, settings=None):
        return self._driver(
            _get_config_request_factory(),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.GetConfig,
            _wrap_get_config_response,
            settings,
        )

    def get_node_labels(self, node_id, settings=None):
        return self._driver(
            _get_node_labels_request_factory(node_id),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.GetNodeLabels,
            _wrap_get_node_labels_response,
            settings,
        )


class DynamicConfigClient(BaseDynamicConfigClient):
    def async_replace_config(self, config, dry_run, allow_unknown_fields, settings=None):
        return self._driver.future(
            _replace_config_request_factory(config, dry_run, allow_unknown_fields),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.ReplaceConfig,
            operation.Operation,
            settings,
        )

    def async_set_config(self, config, dry_run, allow_unknown_fields, settings=None):
        return self._driver.future(
            _set_config_request_factory(config, dry_run, allow_unknown_fields),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.SetConfig,
            operation.Operation,
            settings,
        )

    def async_get_config(self, settings=None):
        return self._driver.future(
            _get_config_request_factory(),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.GetConfig,
            _wrap_get_config_response,
            settings,
        )

    def async_get_node_labels(self, node_id, settings=None):
        return self._driver.future(
            _get_node_labels_request_factory(node_id),
            _apis.DynamicConfigService.Stub,
            _apis.DynamicConfigService.GetNodeLabels,
            _wrap_get_node_labels_response,
            settings,
        )



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureFlag(_message.Message):
    __slots__ = ()
    class Status(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATUS_UNSPECIFIED: _ClassVar[FeatureFlag.Status]
        ENABLED: _ClassVar[FeatureFlag.Status]
        DISABLED: _ClassVar[FeatureFlag.Status]
    STATUS_UNSPECIFIED: FeatureFlag.Status
    ENABLED: FeatureFlag.Status
    DISABLED: FeatureFlag.Status
    def __init__(self) -> None: ...

class CostInfo(_message.Message):
    __slots__ = ("consumed_units",)
    CONSUMED_UNITS_FIELD_NUMBER: _ClassVar[int]
    consumed_units: float
    def __init__(self, consumed_units: _Optional[float] = ...) -> None: ...

class QuotaExceeded(_message.Message):
    __slots__ = ("disk",)
    DISK_FIELD_NUMBER: _ClassVar[int]
    disk: bool
    def __init__(self, disk: bool = ...) -> None: ...

class VirtualTimestamp(_message.Message):
    __slots__ = ("plan_step", "tx_id")
    PLAN_STEP_FIELD_NUMBER: _ClassVar[int]
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    plan_step: int
    tx_id: int
    def __init__(self, plan_step: _Optional[int] = ..., tx_id: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: draft/ydb_keyvalue_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v4.draft.protos import ydb_keyvalue_pb2 as draft_dot_protos_dot_ydb__keyvalue__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x64raft/ydb_keyvalue_v1.proto\x12\x0fYdb.KeyValue.V1\x1a\x1f\x64raft/protos/ydb_keyvalue.proto2\xe6\x07\n\x0fKeyValueService\x12U\n\x0c\x43reateVolume\x12!.Ydb.KeyValue.CreateVolumeRequest\x1a\".Ydb.KeyValue.CreateVolumeResponse\x12O\n\nDropVolume\x12\x1f.Ydb.KeyValue.DropVolumeRequest\x1a .Ydb.KeyValue.DropVolumeResponse\x12R\n\x0b\x41lterVolume\x12 .Ydb.KeyValue.AlterVolumeRequest\x1a!.Ydb.KeyValue.AlterVolumeResponse\x12[\n\x0e\x44\x65scribeVolume\x12#.Ydb.KeyValue.DescribeVolumeRequest\x1a$.Ydb.KeyValue.DescribeVolumeResponse\x12j\n\x13ListLocalPartitions\x12(.Ydb.KeyValue.ListLocalPartitionsRequest\x1a).Ydb.KeyValue.ListLocalPartitionsResponse\x12R\n\x0b\x41\x63quireLock\x12 .Ydb.KeyValue.AcquireLockRequest\x1a!.Ydb.KeyValue.AcquireLockResponse\x12g\n\x12\x45xecuteTransaction\x12\'.Ydb.KeyValue.ExecuteTransactionRequest\x1a(.Ydb.KeyValue.ExecuteTransactionResponse\x12=\n\x04Read\x12\x19.Ydb.KeyValue.ReadRequest\x1a\x1a.Ydb.KeyValue.ReadResponse\x12L\n\tReadRange\x12\x1e.Ydb.KeyValue.ReadRangeRequest\x1a\x1f.Ydb.KeyValue.ReadRangeResponse\x12L\n\tListRange\x12\x1e.Ydb.KeyValue.ListRangeRequest\x1a\x1f.Ydb.KeyValue.ListRangeResponse\x12v\n\x17GetStorageChannelStatus\x12,.Ydb.KeyValue.GetStorageChannelStatusRequest\x1a-.Ydb.KeyValue.GetStorageChannelStatusResponseB^\n\x1dtech.ydb.proto.draft.keyvalueZ=github.com/ydb-platform/ydb-go-genproto/draft/Ydb_KeyValue_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'draft.ydb_keyvalue_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035tech.ydb.proto.draft.keyvalueZ=github.com/ydb-platform/ydb-go-genproto/draft/Ydb_KeyValue_V1'
  _KEYVALUESERVICE._serialized_start=82
  _KEYVALUESERVICE._serialized_end=1080
# @@protoc_insertion_point(module_scope)



import pytest
from ydb.query.session import QuerySession
from ydb.query.pool import QuerySessionPool


@pytest.fixture
def session(driver_sync):
    session = QuerySession(driver_sync)

    yield session

    try:
        session.delete()
    except BaseException:
        pass


@pytest.fixture
def tx(session):
    session.create()
    transaction = session.transaction()

    yield transaction

    try:
        transaction.rollback()
    except BaseException:
        pass


@pytest.fixture
def pool(driver_sync):
    pool = QuerySessionPool(driver_sync)
    yield pool



# -*- coding: utf-8 -*-
import typing

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from .._grpc.v4.draft import (
        ydb_dynamic_config_v1_pb2_grpc,
    )

    from .._grpc.v4.draft.protos import (
        ydb_dynamic_config_pb2,
    )
else:
    from .._grpc.common.draft import (
        ydb_dynamic_config_v1_pb2_grpc,
    )

    try:
        from .._grpc.common.draft.protos import (
            ydb_dynamic_config_pb2,
        )
    except ImportError:
        from .._grpc.common.protos.draft import (
            ydb_dynamic_config_pb2,
        )


ydb_dynamic_config = ydb_dynamic_config_pb2


class DynamicConfigService(object):
    Stub = ydb_dynamic_config_v1_pb2_grpc.DynamicConfigServiceStub

    ReplaceConfig = "ReplaceConfig"
    SetConfig = "SetConfig"
    GetConfig = "GetConfig"
    GetNodeLabels = "GetNodeLabels"



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: draft/ydb_object_storage_v1.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'draft/ydb_object_storage_v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.draft.protos import ydb_object_storage_pb2 as draft_dot_protos_dot_ydb__object__storage__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!draft/ydb_object_storage_v1.proto\x12\x14Ydb.ObjectStorage.V1\x1a%draft/protos/ydb_object_storage.proto2e\n\x14ObjectStorageService\x12M\n\x04List\x12!.Ydb.ObjectStorage.ListingRequest\x1a\".Ydb.ObjectStorage.ListingResponseBl\n&tech.ydb.proto.draft.object_storage.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_ObjectStorage_V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'draft.ydb_object_storage_v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n&tech.ydb.proto.draft.object_storage.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_ObjectStorage_V1'
  _globals['_OBJECTSTORAGESERVICE']._serialized_start=98
  _globals['_OBJECTSTORAGESERVICE']._serialized_end=199
# @@protoc_insertion_point(module_scope)



from protos import ydb_common_pb2 as _ydb_common_pb2
from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MakeDirectoryRequest(_message.Message):
    __slots__ = ("operation_params", "path")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ...) -> None: ...

class MakeDirectoryResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class RemoveDirectoryRequest(_message.Message):
    __slots__ = ("operation_params", "path")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ...) -> None: ...

class RemoveDirectoryResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListDirectoryRequest(_message.Message):
    __slots__ = ("operation_params", "path")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ...) -> None: ...

class ListDirectoryResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class Permissions(_message.Message):
    __slots__ = ("subject", "permission_names")
    SUBJECT_FIELD_NUMBER: _ClassVar[int]
    PERMISSION_NAMES_FIELD_NUMBER: _ClassVar[int]
    subject: str
    permission_names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, subject: _Optional[str] = ..., permission_names: _Optional[_Iterable[str]] = ...) -> None: ...

class Entry(_message.Message):
    __slots__ = ("name", "owner", "type", "effective_permissions", "permissions", "size_bytes", "created_at")
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        TYPE_UNSPECIFIED: _ClassVar[Entry.Type]
        DIRECTORY: _ClassVar[Entry.Type]
        TABLE: _ClassVar[Entry.Type]
        PERS_QUEUE_GROUP: _ClassVar[Entry.Type]
        DATABASE: _ClassVar[Entry.Type]
        RTMR_VOLUME: _ClassVar[Entry.Type]
        BLOCK_STORE_VOLUME: _ClassVar[Entry.Type]
        COORDINATION_NODE: _ClassVar[Entry.Type]
        COLUMN_STORE: _ClassVar[Entry.Type]
        COLUMN_TABLE: _ClassVar[Entry.Type]
        SEQUENCE: _ClassVar[Entry.Type]
        REPLICATION: _ClassVar[Entry.Type]
        TOPIC: _ClassVar[Entry.Type]
        EXTERNAL_TABLE: _ClassVar[Entry.Type]
        EXTERNAL_DATA_SOURCE: _ClassVar[Entry.Type]
        VIEW: _ClassVar[Entry.Type]
        RESOURCE_POOL: _ClassVar[Entry.Type]
        TRANSFER: _ClassVar[Entry.Type]
        SYS_VIEW: _ClassVar[Entry.Type]
    TYPE_UNSPECIFIED: Entry.Type
    DIRECTORY: Entry.Type
    TABLE: Entry.Type
    PERS_QUEUE_GROUP: Entry.Type
    DATABASE: Entry.Type
    RTMR_VOLUME: Entry.Type
    BLOCK_STORE_VOLUME: Entry.Type
    COORDINATION_NODE: Entry.Type
    COLUMN_STORE: Entry.Type
    COLUMN_TABLE: Entry.Type
    SEQUENCE: Entry.Type
    REPLICATION: Entry.Type
    TOPIC: Entry.Type
    EXTERNAL_TABLE: Entry.Type
    EXTERNAL_DATA_SOURCE: Entry.Type
    VIEW: Entry.Type
    RESOURCE_POOL: Entry.Type
    TRANSFER: Entry.Type
    SYS_VIEW: Entry.Type
    NAME_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    EFFECTIVE_PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    name: str
    owner: str
    type: Entry.Type
    effective_permissions: _containers.RepeatedCompositeFieldContainer[Permissions]
    permissions: _containers.RepeatedCompositeFieldContainer[Permissions]
    size_bytes: int
    created_at: _ydb_common_pb2.VirtualTimestamp
    def __init__(self, name: _Optional[str] = ..., owner: _Optional[str] = ..., type: _Optional[_Union[Entry.Type, str]] = ..., effective_permissions: _Optional[_Iterable[_Union[Permissions, _Mapping]]] = ..., permissions: _Optional[_Iterable[_Union[Permissions, _Mapping]]] = ..., size_bytes: _Optional[int] = ..., created_at: _Optional[_Union[_ydb_common_pb2.VirtualTimestamp, _Mapping]] = ...) -> None: ...

class ListDirectoryResult(_message.Message):
    __slots__ = ("self", "children")
    SELF_FIELD_NUMBER: _ClassVar[int]
    CHILDREN_FIELD_NUMBER: _ClassVar[int]
    self: Entry
    children: _containers.RepeatedCompositeFieldContainer[Entry]
    def __init__(self_, self: _Optional[_Union[Entry, _Mapping]] = ..., children: _Optional[_Iterable[_Union[Entry, _Mapping]]] = ...) -> None: ...

class DescribePathRequest(_message.Message):
    __slots__ = ("operation_params", "path")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ...) -> None: ...

class DescribePathResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class DescribePathResult(_message.Message):
    __slots__ = ("self",)
    SELF_FIELD_NUMBER: _ClassVar[int]
    self: Entry
    def __init__(self_, self: _Optional[_Union[Entry, _Mapping]] = ...) -> None: ...

class PermissionsAction(_message.Message):
    __slots__ = ("grant", "revoke", "set", "change_owner")
    GRANT_FIELD_NUMBER: _ClassVar[int]
    REVOKE_FIELD_NUMBER: _ClassVar[int]
    SET_FIELD_NUMBER: _ClassVar[int]
    CHANGE_OWNER_FIELD_NUMBER: _ClassVar[int]
    grant: Permissions
    revoke: Permissions
    set: Permissions
    change_owner: str
    def __init__(self, grant: _Optional[_Union[Permissions, _Mapping]] = ..., revoke: _Optional[_Union[Permissions, _Mapping]] = ..., set: _Optional[_Union[Permissions, _Mapping]] = ..., change_owner: _Optional[str] = ...) -> None: ...

class ModifyPermissionsRequest(_message.Message):
    __slots__ = ("operation_params", "path", "actions", "clear_permissions", "interrupt_inheritance")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    CLEAR_PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    INTERRUPT_INHERITANCE_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    actions: _containers.RepeatedCompositeFieldContainer[PermissionsAction]
    clear_permissions: bool
    interrupt_inheritance: bool
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., actions: _Optional[_Iterable[_Union[PermissionsAction, _Mapping]]] = ..., clear_permissions: bool = ..., interrupt_inheritance: bool = ...) -> None: ...

class ModifyPermissionsResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...



from __future__ import annotations

import abc
import asyncio
import bisect
import enum
from collections import deque
from dataclasses import dataclass, field
import datetime
from typing import Union, Any, List, Dict, Deque, Optional, Tuple

from ydb._grpc.grpcwrapper.ydb_topic import OffsetsRange, Codec
from ydb._topic_reader import topic_reader_asyncio


class ICommittable(abc.ABC):
    @abc.abstractmethod
    def _commit_get_partition_session(self) -> PartitionSession:
        ...

    @abc.abstractmethod
    def _commit_get_offsets_range(self) -> OffsetsRange:
        ...


class ISessionAlive(abc.ABC):
    @property
    @abc.abstractmethod
    def alive(self) -> bool:
        pass


@dataclass
class PublicMessage(ICommittable, ISessionAlive):
    seqno: int
    created_at: datetime.datetime
    message_group_id: str
    session_metadata: Dict[str, str]
    offset: int
    written_at: datetime.datetime
    producer_id: str
    data: Union[bytes, Any]  # set as original decompressed bytes or deserialized object if deserializer set in reader
    metadata_items: Dict[str, bytes]
    _partition_session: PartitionSession
    _commit_start_offset: int
    _commit_end_offset: int

    def _commit_get_partition_session(self) -> PartitionSession:
        return self._partition_session

    def _commit_get_offsets_range(self) -> OffsetsRange:
        return OffsetsRange(self._commit_start_offset, self._commit_end_offset)

    # ISessionAlive implementation
    @property
    def alive(self) -> bool:
        return not self._partition_session.closed

    @property
    def partition_id(self) -> int:
        return self._partition_session.partition_id


@dataclass
class PartitionSession:
    id: int
    state: "PartitionSession.State"
    topic_path: str
    partition_id: int
    committed_offset: int  # last commit offset, acked from server. Processed messages up to the field-1 offset.
    reader_reconnector_id: int
    reader_stream_id: int
    _next_message_start_commit_offset: int = field(init=False)

    # todo: check if deque is optimal
    _ack_waiters: Deque["PartitionSession.CommitAckWaiter"] = field(init=False, default_factory=lambda: deque())

    _state_changed: asyncio.Event = field(init=False, default_factory=lambda: asyncio.Event(), compare=False)

    def __post_init__(self):
        self._next_message_start_commit_offset = self.committed_offset

    def add_waiter(self, end_offset: int) -> "PartitionSession.CommitAckWaiter":
        self._ensure_not_closed()

        waiter = PartitionSession.CommitAckWaiter(end_offset, asyncio.Future())
        if end_offset <= self.committed_offset:
            waiter._finish_ok()
            return waiter

        # fast way
        if self._ack_waiters and self._ack_waiters[-1].end_offset < end_offset:
            self._ack_waiters.append(waiter)
        else:
            bisect.insort(self._ack_waiters, waiter)

        return waiter

    def ack_notify(self, offset: int):
        self._ensure_not_closed()

        self.committed_offset = offset

        if not self._ack_waiters:
            # todo log warning
            # must be never receive ack for not sended request
            return

        while self._ack_waiters:
            if self._ack_waiters[0].end_offset > offset:
                break
            waiter = self._ack_waiters.popleft()
            waiter._finish_ok()

    def _update_last_commited_offset_if_needed(self, offset: int):
        self.committed_offset = max(self.committed_offset, offset)

    def close(self):
        if self.closed:
            return

        self.state = PartitionSession.State.Stopped
        exception = topic_reader_asyncio.PublicTopicReaderPartitionExpiredError()
        for waiter in self._ack_waiters:
            waiter._finish_error(exception)

    @property
    def closed(self):
        return self.state == PartitionSession.State.Stopped

    def end(self):
        if self.closed:
            return

        self.state = PartitionSession.State.Ended

    @property
    def ended(self):
        return self.state == PartitionSession.State.Ended

    def _ensure_not_closed(self):
        if self.state == PartitionSession.State.Stopped:
            raise topic_reader_asyncio.PublicTopicReaderPartitionExpiredError()

    class State(enum.Enum):
        Active = 1
        GracefulShutdown = 2
        Stopped = 3
        Ended = 4

    @dataclass(order=True)
    class CommitAckWaiter:
        end_offset: int
        future: asyncio.Future = field(compare=False)
        _done: bool = field(default=False, init=False)
        _exception: Optional[Exception] = field(default=None, init=False)

        def _finish_ok(self):
            self._done = True
            self.future.set_result(None)

        def _finish_error(self, error: Exception):
            self._exception = error
            self.future.set_exception(error)


@dataclass
class PublicBatch(ICommittable, ISessionAlive):
    messages: List[PublicMessage]
    _partition_session: PartitionSession
    _bytes_size: int
    _codec: Codec

    def _commit_get_partition_session(self) -> PartitionSession:
        return self.messages[0]._commit_get_partition_session()

    def _commit_get_offsets_range(self) -> OffsetsRange:
        return OffsetsRange(
            self.messages[0]._commit_get_offsets_range().start,
            self.messages[-1]._commit_get_offsets_range().end,
        )

    def empty(self) -> bool:
        return len(self.messages) == 0

    # ISessionAlive implementation
    @property
    def alive(self) -> bool:
        return not self._partition_session.closed

    def pop_message(self) -> PublicMessage:
        return self.messages.pop(0)

    def _extend(self, batch: PublicBatch) -> None:
        self.messages.extend(batch.messages)
        self._bytes_size += batch._bytes_size

    def _pop(self) -> Tuple[List[PublicMessage], bool]:
        msgs_left = True if len(self.messages) > 1 else False
        return self.messages.pop(0), msgs_left

    def _pop_batch(self, message_count: int) -> PublicBatch:
        initial_length = len(self.messages)

        if message_count >= initial_length:
            raise ValueError("Pop batch with size >= actual size is not supported.")

        one_message_size = self._bytes_size // initial_length

        new_batch = PublicBatch(
            messages=self.messages[:message_count],
            _partition_session=self._partition_session,
            _bytes_size=one_message_size * message_count,
            _codec=self._codec,
        )

        self.messages = self.messages[message_count:]
        self._bytes_size = self._bytes_size - new_batch._bytes_size

        return new_batch

    def _update_partition_offsets(self, tx, exc=None):
        if exc is not None:
            return
        offsets = self._commit_get_offsets_range()
        self._partition_session._update_last_commited_offset_if_needed(offsets.end)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ydb_import_v1.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'ydb_import_v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_import_pb2 as protos_dot_ydb__import__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13ydb_import_v1.proto\x12\rYdb.Import.V1\x1a\x17protos/ydb_import.proto2\xde\x03\n\rImportService\x12Q\n\x0cImportFromS3\x12\x1f.Ydb.Import.ImportFromS3Request\x1a .Ydb.Import.ImportFromS3Response\x12Q\n\x0cImportFromFs\x12\x1f.Ydb.Import.ImportFromFsRequest\x1a .Ydb.Import.ImportFromFsResponse\x12l\n\x15ListObjectsInS3Export\x12(.Ydb.Import.ListObjectsInS3ExportRequest\x1a).Ydb.Import.ListObjectsInS3ExportResponse\x12l\n\x15ListObjectsInFsExport\x12(.Ydb.Import.ListObjectsInFsExportRequest\x1a).Ydb.Import.ListObjectsInFsExportResponse\x12K\n\nImportData\x12\x1d.Ydb.Import.ImportDataRequest\x1a\x1e.Ydb.Import.ImportDataResponseBR\n\x19tech.ydb.proto.import_.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Import_V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_import_v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\031tech.ydb.proto.import_.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Import_V1'
  _globals['_IMPORTSERVICE']._serialized_start=64
  _globals['_IMPORTSERVICE']._serialized_end=542
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_scheme_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v4.protos import ydb_scheme_pb2 as protos_dot_ydb__scheme__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13ydb_scheme_v1.proto\x12\rYdb.Scheme.V1\x1a\x17protos/ydb_scheme.proto2\xcc\x03\n\rSchemeService\x12T\n\rMakeDirectory\x12 .Ydb.Scheme.MakeDirectoryRequest\x1a!.Ydb.Scheme.MakeDirectoryResponse\x12Z\n\x0fRemoveDirectory\x12\".Ydb.Scheme.RemoveDirectoryRequest\x1a#.Ydb.Scheme.RemoveDirectoryResponse\x12T\n\rListDirectory\x12 .Ydb.Scheme.ListDirectoryRequest\x1a!.Ydb.Scheme.ListDirectoryResponse\x12Q\n\x0c\x44\x65scribePath\x12\x1f.Ydb.Scheme.DescribePathRequest\x1a .Ydb.Scheme.DescribePathResponse\x12`\n\x11ModifyPermissions\x12$.Ydb.Scheme.ModifyPermissionsRequest\x1a%.Ydb.Scheme.ModifyPermissionsResponseBQ\n\x18tech.ydb.proto.scheme.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Scheme_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_scheme_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030tech.ydb.proto.scheme.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Scheme_V1'
  _SCHEMESERVICE._serialized_start=64
  _SCHEMESERVICE._serialized_end=524
# @@protoc_insertion_point(module_scope)



import pytest

import ydb


def test_close_basic_logic_case_1(driver_sync):
    pool = ydb.SessionPool(driver_sync, 1)
    s = pool.acquire()

    pool.stop()

    with pytest.raises(ValueError):
        pool.acquire()

    pool.release(s)
    assert pool._pool_impl._active_count == 0


def test_close_basic_logic_case_2(driver_sync):
    pool = ydb.SessionPool(driver_sync, 10)
    acquired = []

    for _ in range(10):
        acquired.append(pool.acquire())

    for _ in range(3):
        pool.release(acquired.pop(-1))

    pool.stop()
    assert pool._pool_impl._active_count == 7

    while acquired:
        pool.release(acquired.pop(-1))

    assert pool._pool_impl._active_count == 0

    with pytest.raises(ValueError):
        pool.acquire()

    pool.stop()



from protos import ydb_operation_pb2 as _ydb_operation_pb2
from protos import ydb_query_pb2 as _ydb_query_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor



import datetime

from google.protobuf import duration_pb2 as _duration_pb2
from protos.annotations import validation_pb2 as _validation_pb2
from protos import ydb_issue_message_pb2 as _ydb_issue_message_pb2
from protos import ydb_operation_pb2 as _ydb_operation_pb2
from protos import ydb_query_stats_pb2 as _ydb_query_stats_pb2
from protos import ydb_status_codes_pb2 as _ydb_status_codes_pb2
from protos import ydb_formats_pb2 as _ydb_formats_pb2
from protos import ydb_value_pb2 as _ydb_value_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Syntax(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SYNTAX_UNSPECIFIED: _ClassVar[Syntax]
    SYNTAX_YQL_V1: _ClassVar[Syntax]
    SYNTAX_PG: _ClassVar[Syntax]

class ExecMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    EXEC_MODE_UNSPECIFIED: _ClassVar[ExecMode]
    EXEC_MODE_PARSE: _ClassVar[ExecMode]
    EXEC_MODE_VALIDATE: _ClassVar[ExecMode]
    EXEC_MODE_EXPLAIN: _ClassVar[ExecMode]
    EXEC_MODE_EXECUTE: _ClassVar[ExecMode]

class StatsMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    STATS_MODE_UNSPECIFIED: _ClassVar[StatsMode]
    STATS_MODE_NONE: _ClassVar[StatsMode]
    STATS_MODE_BASIC: _ClassVar[StatsMode]
    STATS_MODE_FULL: _ClassVar[StatsMode]
    STATS_MODE_PROFILE: _ClassVar[StatsMode]

class SchemaInclusionMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SCHEMA_INCLUSION_MODE_UNSPECIFIED: _ClassVar[SchemaInclusionMode]
    SCHEMA_INCLUSION_MODE_ALWAYS: _ClassVar[SchemaInclusionMode]
    SCHEMA_INCLUSION_MODE_FIRST_ONLY: _ClassVar[SchemaInclusionMode]

class ExecStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    EXEC_STATUS_UNSPECIFIED: _ClassVar[ExecStatus]
    EXEC_STATUS_STARTING: _ClassVar[ExecStatus]
    EXEC_STATUS_ABORTED: _ClassVar[ExecStatus]
    EXEC_STATUS_CANCELLED: _ClassVar[ExecStatus]
    EXEC_STATUS_COMPLETED: _ClassVar[ExecStatus]
    EXEC_STATUS_FAILED: _ClassVar[ExecStatus]
SYNTAX_UNSPECIFIED: Syntax
SYNTAX_YQL_V1: Syntax
SYNTAX_PG: Syntax
EXEC_MODE_UNSPECIFIED: ExecMode
EXEC_MODE_PARSE: ExecMode
EXEC_MODE_VALIDATE: ExecMode
EXEC_MODE_EXPLAIN: ExecMode
EXEC_MODE_EXECUTE: ExecMode
STATS_MODE_UNSPECIFIED: StatsMode
STATS_MODE_NONE: StatsMode
STATS_MODE_BASIC: StatsMode
STATS_MODE_FULL: StatsMode
STATS_MODE_PROFILE: StatsMode
SCHEMA_INCLUSION_MODE_UNSPECIFIED: SchemaInclusionMode
SCHEMA_INCLUSION_MODE_ALWAYS: SchemaInclusionMode
SCHEMA_INCLUSION_MODE_FIRST_ONLY: SchemaInclusionMode
EXEC_STATUS_UNSPECIFIED: ExecStatus
EXEC_STATUS_STARTING: ExecStatus
EXEC_STATUS_ABORTED: ExecStatus
EXEC_STATUS_CANCELLED: ExecStatus
EXEC_STATUS_COMPLETED: ExecStatus
EXEC_STATUS_FAILED: ExecStatus

class CreateSessionRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class CreateSessionResponse(_message.Message):
    __slots__ = ("status", "issues", "session_id", "node_id")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    session_id: str
    node_id: int
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., session_id: _Optional[str] = ..., node_id: _Optional[int] = ...) -> None: ...

class DeleteSessionRequest(_message.Message):
    __slots__ = ("session_id",)
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class DeleteSessionResponse(_message.Message):
    __slots__ = ("status", "issues")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class AttachSessionRequest(_message.Message):
    __slots__ = ("session_id",)
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class SessionState(_message.Message):
    __slots__ = ("status", "issues")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class SerializableModeSettings(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class OnlineModeSettings(_message.Message):
    __slots__ = ("allow_inconsistent_reads",)
    ALLOW_INCONSISTENT_READS_FIELD_NUMBER: _ClassVar[int]
    allow_inconsistent_reads: bool
    def __init__(self, allow_inconsistent_reads: bool = ...) -> None: ...

class StaleModeSettings(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SnapshotModeSettings(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SnapshotRWModeSettings(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class TransactionSettings(_message.Message):
    __slots__ = ("serializable_read_write", "online_read_only", "stale_read_only", "snapshot_read_only", "snapshot_read_write")
    SERIALIZABLE_READ_WRITE_FIELD_NUMBER: _ClassVar[int]
    ONLINE_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    STALE_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    SNAPSHOT_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    SNAPSHOT_READ_WRITE_FIELD_NUMBER: _ClassVar[int]
    serializable_read_write: SerializableModeSettings
    online_read_only: OnlineModeSettings
    stale_read_only: StaleModeSettings
    snapshot_read_only: SnapshotModeSettings
    snapshot_read_write: SnapshotRWModeSettings
    def __init__(self, serializable_read_write: _Optional[_Union[SerializableModeSettings, _Mapping]] = ..., online_read_only: _Optional[_Union[OnlineModeSettings, _Mapping]] = ..., stale_read_only: _Optional[_Union[StaleModeSettings, _Mapping]] = ..., snapshot_read_only: _Optional[_Union[SnapshotModeSettings, _Mapping]] = ..., snapshot_read_write: _Optional[_Union[SnapshotRWModeSettings, _Mapping]] = ...) -> None: ...

class TransactionControl(_message.Message):
    __slots__ = ("tx_id", "begin_tx", "commit_tx")
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    BEGIN_TX_FIELD_NUMBER: _ClassVar[int]
    COMMIT_TX_FIELD_NUMBER: _ClassVar[int]
    tx_id: str
    begin_tx: TransactionSettings
    commit_tx: bool
    def __init__(self, tx_id: _Optional[str] = ..., begin_tx: _Optional[_Union[TransactionSettings, _Mapping]] = ..., commit_tx: bool = ...) -> None: ...

class BeginTransactionRequest(_message.Message):
    __slots__ = ("session_id", "tx_settings")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TX_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    tx_settings: TransactionSettings
    def __init__(self, session_id: _Optional[str] = ..., tx_settings: _Optional[_Union[TransactionSettings, _Mapping]] = ...) -> None: ...

class TransactionMeta(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class BeginTransactionResponse(_message.Message):
    __slots__ = ("status", "issues", "tx_meta")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    TX_META_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    tx_meta: TransactionMeta
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., tx_meta: _Optional[_Union[TransactionMeta, _Mapping]] = ...) -> None: ...

class CommitTransactionRequest(_message.Message):
    __slots__ = ("session_id", "tx_id")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    tx_id: str
    def __init__(self, session_id: _Optional[str] = ..., tx_id: _Optional[str] = ...) -> None: ...

class CommitTransactionResponse(_message.Message):
    __slots__ = ("status", "issues")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class RollbackTransactionRequest(_message.Message):
    __slots__ = ("session_id", "tx_id")
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    tx_id: str
    def __init__(self, session_id: _Optional[str] = ..., tx_id: _Optional[str] = ...) -> None: ...

class RollbackTransactionResponse(_message.Message):
    __slots__ = ("status", "issues")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class QueryContent(_message.Message):
    __slots__ = ("syntax", "text")
    SYNTAX_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    syntax: Syntax
    text: str
    def __init__(self, syntax: _Optional[_Union[Syntax, str]] = ..., text: _Optional[str] = ...) -> None: ...

class ExecuteQueryRequest(_message.Message):
    __slots__ = ("session_id", "exec_mode", "tx_control", "query_content", "parameters", "stats_mode", "concurrent_result_sets", "response_part_limit_bytes", "pool_id", "stats_period_ms", "schema_inclusion_mode", "result_set_format", "arrow_format_settings")
    class ParametersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _ydb_value_pb2.TypedValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_ydb_value_pb2.TypedValue, _Mapping]] = ...) -> None: ...
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    EXEC_MODE_FIELD_NUMBER: _ClassVar[int]
    TX_CONTROL_FIELD_NUMBER: _ClassVar[int]
    QUERY_CONTENT_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    STATS_MODE_FIELD_NUMBER: _ClassVar[int]
    CONCURRENT_RESULT_SETS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_PART_LIMIT_BYTES_FIELD_NUMBER: _ClassVar[int]
    POOL_ID_FIELD_NUMBER: _ClassVar[int]
    STATS_PERIOD_MS_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_INCLUSION_MODE_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_FORMAT_FIELD_NUMBER: _ClassVar[int]
    ARROW_FORMAT_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    exec_mode: ExecMode
    tx_control: TransactionControl
    query_content: QueryContent
    parameters: _containers.MessageMap[str, _ydb_value_pb2.TypedValue]
    stats_mode: StatsMode
    concurrent_result_sets: bool
    response_part_limit_bytes: int
    pool_id: str
    stats_period_ms: int
    schema_inclusion_mode: SchemaInclusionMode
    result_set_format: _ydb_value_pb2.ResultSet.Format
    arrow_format_settings: _ydb_formats_pb2.ArrowFormatSettings
    def __init__(self, session_id: _Optional[str] = ..., exec_mode: _Optional[_Union[ExecMode, str]] = ..., tx_control: _Optional[_Union[TransactionControl, _Mapping]] = ..., query_content: _Optional[_Union[QueryContent, _Mapping]] = ..., parameters: _Optional[_Mapping[str, _ydb_value_pb2.TypedValue]] = ..., stats_mode: _Optional[_Union[StatsMode, str]] = ..., concurrent_result_sets: bool = ..., response_part_limit_bytes: _Optional[int] = ..., pool_id: _Optional[str] = ..., stats_period_ms: _Optional[int] = ..., schema_inclusion_mode: _Optional[_Union[SchemaInclusionMode, str]] = ..., result_set_format: _Optional[_Union[_ydb_value_pb2.ResultSet.Format, str]] = ..., arrow_format_settings: _Optional[_Union[_ydb_formats_pb2.ArrowFormatSettings, _Mapping]] = ...) -> None: ...

class ResultSetMeta(_message.Message):
    __slots__ = ("columns",)
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    columns: _containers.RepeatedCompositeFieldContainer[_ydb_value_pb2.Column]
    def __init__(self, columns: _Optional[_Iterable[_Union[_ydb_value_pb2.Column, _Mapping]]] = ...) -> None: ...

class ExecuteQueryResponsePart(_message.Message):
    __slots__ = ("status", "issues", "result_set_index", "result_set", "exec_stats", "tx_meta")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_INDEX_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_FIELD_NUMBER: _ClassVar[int]
    EXEC_STATS_FIELD_NUMBER: _ClassVar[int]
    TX_META_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    result_set_index: int
    result_set: _ydb_value_pb2.ResultSet
    exec_stats: _ydb_query_stats_pb2.QueryStats
    tx_meta: TransactionMeta
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., result_set_index: _Optional[int] = ..., result_set: _Optional[_Union[_ydb_value_pb2.ResultSet, _Mapping]] = ..., exec_stats: _Optional[_Union[_ydb_query_stats_pb2.QueryStats, _Mapping]] = ..., tx_meta: _Optional[_Union[TransactionMeta, _Mapping]] = ...) -> None: ...

class ExecuteScriptRequest(_message.Message):
    __slots__ = ("operation_params", "exec_mode", "script_content", "parameters", "stats_mode", "results_ttl", "pool_id")
    class ParametersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _ydb_value_pb2.TypedValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_ydb_value_pb2.TypedValue, _Mapping]] = ...) -> None: ...
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    EXEC_MODE_FIELD_NUMBER: _ClassVar[int]
    SCRIPT_CONTENT_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    STATS_MODE_FIELD_NUMBER: _ClassVar[int]
    RESULTS_TTL_FIELD_NUMBER: _ClassVar[int]
    POOL_ID_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    exec_mode: ExecMode
    script_content: QueryContent
    parameters: _containers.MessageMap[str, _ydb_value_pb2.TypedValue]
    stats_mode: StatsMode
    results_ttl: _duration_pb2.Duration
    pool_id: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., exec_mode: _Optional[_Union[ExecMode, str]] = ..., script_content: _Optional[_Union[QueryContent, _Mapping]] = ..., parameters: _Optional[_Mapping[str, _ydb_value_pb2.TypedValue]] = ..., stats_mode: _Optional[_Union[StatsMode, str]] = ..., results_ttl: _Optional[_Union[datetime.timedelta, _duration_pb2.Duration, _Mapping]] = ..., pool_id: _Optional[str] = ...) -> None: ...

class ExecuteScriptMetadata(_message.Message):
    __slots__ = ("execution_id", "exec_status", "script_content", "result_sets_meta", "exec_mode", "exec_stats")
    EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    EXEC_STATUS_FIELD_NUMBER: _ClassVar[int]
    SCRIPT_CONTENT_FIELD_NUMBER: _ClassVar[int]
    RESULT_SETS_META_FIELD_NUMBER: _ClassVar[int]
    EXEC_MODE_FIELD_NUMBER: _ClassVar[int]
    EXEC_STATS_FIELD_NUMBER: _ClassVar[int]
    execution_id: str
    exec_status: ExecStatus
    script_content: QueryContent
    result_sets_meta: _containers.RepeatedCompositeFieldContainer[ResultSetMeta]
    exec_mode: ExecMode
    exec_stats: _ydb_query_stats_pb2.QueryStats
    def __init__(self, execution_id: _Optional[str] = ..., exec_status: _Optional[_Union[ExecStatus, str]] = ..., script_content: _Optional[_Union[QueryContent, _Mapping]] = ..., result_sets_meta: _Optional[_Iterable[_Union[ResultSetMeta, _Mapping]]] = ..., exec_mode: _Optional[_Union[ExecMode, str]] = ..., exec_stats: _Optional[_Union[_ydb_query_stats_pb2.QueryStats, _Mapping]] = ...) -> None: ...

class FetchScriptResultsRequest(_message.Message):
    __slots__ = ("operation_id", "result_set_index", "fetch_token", "rows_limit")
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_INDEX_FIELD_NUMBER: _ClassVar[int]
    FETCH_TOKEN_FIELD_NUMBER: _ClassVar[int]
    ROWS_LIMIT_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    result_set_index: int
    fetch_token: str
    rows_limit: int
    def __init__(self, operation_id: _Optional[str] = ..., result_set_index: _Optional[int] = ..., fetch_token: _Optional[str] = ..., rows_limit: _Optional[int] = ...) -> None: ...

class FetchScriptResultsResponse(_message.Message):
    __slots__ = ("status", "issues", "result_set_index", "result_set", "next_fetch_token")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_INDEX_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_FIELD_NUMBER: _ClassVar[int]
    NEXT_FETCH_TOKEN_FIELD_NUMBER: _ClassVar[int]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    result_set_index: int
    result_set: _ydb_value_pb2.ResultSet
    next_fetch_token: str
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., result_set_index: _Optional[int] = ..., result_set: _Optional[_Union[_ydb_value_pb2.ResultSet, _Mapping]] = ..., next_fetch_token: _Optional[str] = ...) -> None: ...

class Script(_message.Message):
    __slots__ = ("script_content",)
    SCRIPT_CONTENT_FIELD_NUMBER: _ClassVar[int]
    script_content: QueryContent
    def __init__(self, script_content: _Optional[_Union[QueryContent, _Mapping]] = ...) -> None: ...



class Warning(Exception):
    pass


class Error(Exception):
    def __init__(self, message, issues=None, status=None):

        pretty_issues = _pretty_issues(issues)
        message = message if pretty_issues is None else pretty_issues

        super(Error, self).__init__(message)
        self.issues = issues
        self.message = message
        self.status = status


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


def _pretty_issues(issues):
    if issues is None:
        return None

    children_messages = [_get_messages(issue, root=True) for issue in issues]

    if None in children_messages:
        return None

    return "\n" + "\n".join(children_messages)


def _get_messages(issue, max_depth=100, indent=2, depth=0, root=False):
    if depth >= max_depth:
        return None
    margin_str = " " * depth * indent
    pre_message = ""
    children = ""
    if issue.issues:
        collapsed_messages = []
        while not root and len(issue.issues) == 1:
            collapsed_messages.append(issue.message)
            issue = issue.issues[0]
        if collapsed_messages:
            pre_message = margin_str + ", ".join(collapsed_messages) + "\n"
            depth += 1
            margin_str = " " * depth * indent
        else:
            pre_message = ""

        children_messages = [
            _get_messages(iss, max_depth=max_depth, indent=indent, depth=depth + 1) for iss in issue.issues
        ]

        if None in children_messages:
            return None

        children = "\n".join(children_messages)

    return (
        pre_message
        + margin_str
        + issue.message
        + "\n"
        + margin_str
        + "severity level: "
        + str(issue.severity)
        + "\n"
        + margin_str
        + "issue code: "
        + str(issue.issue_code)
        + "\n"
        + children
    )



# -*- coding: utf-8 -*-
import typing

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from ._grpc.v4 import (
        ydb_cms_v1_pb2_grpc,
        ydb_discovery_v1_pb2_grpc,
        ydb_scheme_v1_pb2_grpc,
        ydb_table_v1_pb2_grpc,
        ydb_operation_v1_pb2_grpc,
        ydb_topic_v1_pb2_grpc,
        ydb_query_v1_pb2_grpc,
        ydb_coordination_v1_pb2_grpc,
    )

    from ._grpc.v4.protos import (
        ydb_status_codes_pb2,
        ydb_discovery_pb2,
        ydb_scheme_pb2,
        ydb_table_pb2,
        ydb_value_pb2,
        ydb_operation_pb2,
        ydb_common_pb2,
        ydb_query_pb2,
        ydb_coordination_pb2,
    )

else:
    from ._grpc.common import (
        ydb_cms_v1_pb2_grpc,
        ydb_discovery_v1_pb2_grpc,
        ydb_scheme_v1_pb2_grpc,
        ydb_table_v1_pb2_grpc,
        ydb_operation_v1_pb2_grpc,
        ydb_topic_v1_pb2_grpc,
        ydb_query_v1_pb2_grpc,
        ydb_coordination_v1_pb2_grpc,
    )

    from ._grpc.common.protos import (
        ydb_status_codes_pb2,
        ydb_discovery_pb2,
        ydb_scheme_pb2,
        ydb_table_pb2,
        ydb_value_pb2,
        ydb_operation_pb2,
        ydb_common_pb2,
        ydb_query_pb2,
        ydb_coordination_pb2,
    )


StatusIds = ydb_status_codes_pb2.StatusIds
FeatureFlag = ydb_common_pb2.FeatureFlag
primitive_types = ydb_value_pb2.Type.PrimitiveTypeId
ydb_value = ydb_value_pb2
ydb_scheme = ydb_scheme_pb2
ydb_table = ydb_table_pb2
ydb_discovery = ydb_discovery_pb2
ydb_operation = ydb_operation_pb2
ydb_query = ydb_query_pb2
ydb_coordination = ydb_coordination_pb2


class CmsService(object):
    Stub = ydb_cms_v1_pb2_grpc.CmsServiceStub


class DiscoveryService(object):
    Stub = ydb_discovery_v1_pb2_grpc.DiscoveryServiceStub
    ListEndpoints = "ListEndpoints"


class OperationService(object):
    Stub = ydb_operation_v1_pb2_grpc.OperationServiceStub
    ForgetOperation = "ForgetOperation"
    GetOperation = "GetOperation"
    CancelOperation = "CancelOperation"


class SchemeService(object):
    Stub = ydb_scheme_v1_pb2_grpc.SchemeServiceStub
    MakeDirectory = "MakeDirectory"
    RemoveDirectory = "RemoveDirectory"
    ListDirectory = "ListDirectory"
    DescribePath = "DescribePath"
    ModifyPermissions = "ModifyPermissions"


class TableService(object):
    Stub = ydb_table_v1_pb2_grpc.TableServiceStub

    StreamExecuteScanQuery = "StreamExecuteScanQuery"
    ExplainDataQuery = "ExplainDataQuery"
    CreateTable = "CreateTable"
    DropTable = "DropTable"
    AlterTable = "AlterTable"
    CopyTables = "CopyTables"
    RenameTables = "RenameTables"
    DescribeTable = "DescribeTable"
    CreateSession = "CreateSession"
    DeleteSession = "DeleteSession"
    ExecuteSchemeQuery = "ExecuteSchemeQuery"
    PrepareDataQuery = "PrepareDataQuery"
    ExecuteDataQuery = "ExecuteDataQuery"
    BeginTransaction = "BeginTransaction"
    CommitTransaction = "CommitTransaction"
    RollbackTransaction = "RollbackTransaction"
    KeepAlive = "KeepAlive"
    StreamReadTable = "StreamReadTable"
    BulkUpsert = "BulkUpsert"


class TopicService(object):
    Stub = ydb_topic_v1_pb2_grpc.TopicServiceStub

    CreateTopic = "CreateTopic"
    DescribeTopic = "DescribeTopic"
    AlterTopic = "AlterTopic"
    DropTopic = "DropTopic"
    StreamRead = "StreamRead"
    StreamWrite = "StreamWrite"
    UpdateOffsetsInTransaction = "UpdateOffsetsInTransaction"
    CommitOffset = "CommitOffset"


class QueryService(object):
    Stub = ydb_query_v1_pb2_grpc.QueryServiceStub

    CreateSession = "CreateSession"
    DeleteSession = "DeleteSession"
    AttachSession = "AttachSession"

    BeginTransaction = "BeginTransaction"
    CommitTransaction = "CommitTransaction"
    RollbackTransaction = "RollbackTransaction"

    ExecuteQuery = "ExecuteQuery"
    ExecuteScript = "ExecuteScript"
    FetchScriptResults = "FetchScriptResults"


class CoordinationService(object):
    Stub = ydb_coordination_v1_pb2_grpc.CoordinationServiceStub
    CreateNode = "CreateNode"
    AlterNode = "AlterNode"
    DropNode = "DropNode"
    DescribeNode = "DescribeNode"
    SessionRequest = "SessionRequest"
    Session = "Session"



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_topic_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_topic_pb2 as protos_dot_ydb__topic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12ydb_topic_v1.proto\x12\x0cYdb.Topic.V1\x1a\x16protos/ydb_topic.proto2\xb7\x06\n\x0cTopicService\x12\x65\n\x0bStreamWrite\x12(.Ydb.Topic.StreamWriteMessage.FromClient\x1a(.Ydb.Topic.StreamWriteMessage.FromServer(\x01\x30\x01\x12\x62\n\nStreamRead\x12\'.Ydb.Topic.StreamReadMessage.FromClient\x1a\'.Ydb.Topic.StreamReadMessage.FromServer(\x01\x30\x01\x12O\n\x0c\x43ommitOffset\x12\x1e.Ydb.Topic.CommitOffsetRequest\x1a\x1f.Ydb.Topic.CommitOffsetResponse\x12y\n\x1aUpdateOffsetsInTransaction\x12,.Ydb.Topic.UpdateOffsetsInTransactionRequest\x1a-.Ydb.Topic.UpdateOffsetsInTransactionResponse\x12L\n\x0b\x43reateTopic\x12\x1d.Ydb.Topic.CreateTopicRequest\x1a\x1e.Ydb.Topic.CreateTopicResponse\x12R\n\rDescribeTopic\x12\x1f.Ydb.Topic.DescribeTopicRequest\x1a .Ydb.Topic.DescribeTopicResponse\x12[\n\x10\x44\x65scribeConsumer\x12\".Ydb.Topic.DescribeConsumerRequest\x1a#.Ydb.Topic.DescribeConsumerResponse\x12I\n\nAlterTopic\x12\x1c.Ydb.Topic.AlterTopicRequest\x1a\x1d.Ydb.Topic.AlterTopicResponse\x12\x46\n\tDropTopic\x12\x1b.Ydb.Topic.DropTopicRequest\x1a\x1c.Ydb.Topic.DropTopicResponseBR\n\x17tech.ydb.proto.topic.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Topic_V1\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_topic_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\027tech.ydb.proto.topic.v1Z4github.com/ydb-platform/ydb-go-genproto/Ydb_Topic_V1\370\001\001'
  _TOPICSERVICE._serialized_start=61
  _TOPICSERVICE._serialized_end=884
# @@protoc_insertion_point(module_scope)



from __future__ import absolute_import
from __future__ import unicode_literals

from .connection import Connection
from .errors import (
    Warning,
    Error,
    InterfaceError,
    DatabaseError,
    DataError,
    OperationalError,
    IntegrityError,
    InternalError,
    ProgrammingError,
    NotSupportedError,
)

version = "0.0.31"

version_info = (
    1,
    0,
    0,
)

apilevel = "1.0"

threadsafety = 0

paramstyle = "qmark"

errors = (
    Warning,
    Error,
    InterfaceError,
    DatabaseError,
    DataError,
    OperationalError,
    IntegrityError,
    InternalError,
    ProgrammingError,
    NotSupportedError,
)


def connect(*args, **kwargs):
    return Connection(*args, **kwargs)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class OperationStats(_message.Message):
    __slots__ = ("rows", "bytes")
    ROWS_FIELD_NUMBER: _ClassVar[int]
    BYTES_FIELD_NUMBER: _ClassVar[int]
    rows: int
    bytes: int
    def __init__(self, rows: _Optional[int] = ..., bytes: _Optional[int] = ...) -> None: ...

class TableAccessStats(_message.Message):
    __slots__ = ("name", "reads", "updates", "deletes", "partitions_count")
    NAME_FIELD_NUMBER: _ClassVar[int]
    READS_FIELD_NUMBER: _ClassVar[int]
    UPDATES_FIELD_NUMBER: _ClassVar[int]
    DELETES_FIELD_NUMBER: _ClassVar[int]
    PARTITIONS_COUNT_FIELD_NUMBER: _ClassVar[int]
    name: str
    reads: OperationStats
    updates: OperationStats
    deletes: OperationStats
    partitions_count: int
    def __init__(self, name: _Optional[str] = ..., reads: _Optional[_Union[OperationStats, _Mapping]] = ..., updates: _Optional[_Union[OperationStats, _Mapping]] = ..., deletes: _Optional[_Union[OperationStats, _Mapping]] = ..., partitions_count: _Optional[int] = ...) -> None: ...

class QueryPhaseStats(_message.Message):
    __slots__ = ("duration_us", "table_access", "cpu_time_us", "affected_shards", "literal_phase")
    DURATION_US_FIELD_NUMBER: _ClassVar[int]
    TABLE_ACCESS_FIELD_NUMBER: _ClassVar[int]
    CPU_TIME_US_FIELD_NUMBER: _ClassVar[int]
    AFFECTED_SHARDS_FIELD_NUMBER: _ClassVar[int]
    LITERAL_PHASE_FIELD_NUMBER: _ClassVar[int]
    duration_us: int
    table_access: _containers.RepeatedCompositeFieldContainer[TableAccessStats]
    cpu_time_us: int
    affected_shards: int
    literal_phase: bool
    def __init__(self, duration_us: _Optional[int] = ..., table_access: _Optional[_Iterable[_Union[TableAccessStats, _Mapping]]] = ..., cpu_time_us: _Optional[int] = ..., affected_shards: _Optional[int] = ..., literal_phase: bool = ...) -> None: ...

class CompilationStats(_message.Message):
    __slots__ = ("from_cache", "duration_us", "cpu_time_us")
    FROM_CACHE_FIELD_NUMBER: _ClassVar[int]
    DURATION_US_FIELD_NUMBER: _ClassVar[int]
    CPU_TIME_US_FIELD_NUMBER: _ClassVar[int]
    from_cache: bool
    duration_us: int
    cpu_time_us: int
    def __init__(self, from_cache: bool = ..., duration_us: _Optional[int] = ..., cpu_time_us: _Optional[int] = ...) -> None: ...

class QueryStats(_message.Message):
    __slots__ = ("query_phases", "compilation", "process_cpu_time_us", "query_plan", "query_ast", "total_duration_us", "total_cpu_time_us")
    QUERY_PHASES_FIELD_NUMBER: _ClassVar[int]
    COMPILATION_FIELD_NUMBER: _ClassVar[int]
    PROCESS_CPU_TIME_US_FIELD_NUMBER: _ClassVar[int]
    QUERY_PLAN_FIELD_NUMBER: _ClassVar[int]
    QUERY_AST_FIELD_NUMBER: _ClassVar[int]
    TOTAL_DURATION_US_FIELD_NUMBER: _ClassVar[int]
    TOTAL_CPU_TIME_US_FIELD_NUMBER: _ClassVar[int]
    query_phases: _containers.RepeatedCompositeFieldContainer[QueryPhaseStats]
    compilation: CompilationStats
    process_cpu_time_us: int
    query_plan: str
    query_ast: str
    total_duration_us: int
    total_cpu_time_us: int
    def __init__(self, query_phases: _Optional[_Iterable[_Union[QueryPhaseStats, _Mapping]]] = ..., compilation: _Optional[_Union[CompilationStats, _Mapping]] = ..., process_cpu_time_us: _Optional[int] = ..., query_plan: _Optional[str] = ..., query_ast: _Optional[str] = ..., total_duration_us: _Optional[int] = ..., total_cpu_time_us: _Optional[int] = ...) -> None: ...
